from baseten.models import flan_t5, llama, stable_diffusion, whisper

StableDiffusionPipeline = stable_diffusion.StableDiffusionPipeline
FlanT5 = flan_t5.FlanT5
Whisper = whisper.Whisper
Llama = llama.Llama
